import json
import os
import sys
from glob import glob
import botocore
import click
import pandas as pd
import tsdataformat
from seaflowpy import cloud
from seaflowpy import db
from seaflowpy import errors as sfperrors
from seaflowpy import fileio
from seaflowpy import seaflowfile
from seaflowpy import sfl



@click.group()
def sfl_cmd():
    """SFL file operations subcommand."""
    pass


@sfl_cmd.command('convert-gga')
@click.argument('sfl-file', nargs=1, type=click.File())
def sfl_convert_gga_cmd(sfl_file):
    """
    Converts GGA coords to decimal degrees.

    To read from STDIN use '-' for SFL_FILE. Prints modified SFL file to STDOUT.
    """
    df = sfl.read_file(sfl_file, convert_numerics=False)
    try:
        df = sfl.convert_gga2dd(df)
    except ValueError as e:
        raise click.ClickException(str(e))
    sfl.save_to_file(df, sys.stdout)


@sfl_cmd.command('detect-gga')
@click.argument('sfl-file', nargs=1, type=click.File())
def sfl_detect_gga_cmd(sfl_file):
    """
    Detects GGA coordinates in SFL_FILE.

    To read from STDIN use '-' for SFL_FILE. Prints 'True' to STDOUT if any GGA
    coordinates are found, else 'False'.
    """
    df = sfl.read_file(sfl_file, convert_numerics=False)
    # Has any GGA coordinates?
    if sfl.has_gga(df):
        click.echo('True')
    else:
        click.echo('False')


@sfl_cmd.command('dedup')
@click.argument('sfl-file', nargs=1, type=click.File())
def sfl_dedup_cmd(sfl_file):
    """
    Removes duplicate 'FILE' lines.

    To read from STDIN use '-' for SFL_FILE. Removes lines with duplicate file
    entries and prints modified SFL to STDOUT. Because it's impossible to know
    which of the duplicated SFL entries corresponds to which EVT file, all
    duplicate rows are removed. Prints a unique list of files removed to STDERR.
    Duplicate files should also be removed from EVT data sets.
    """
    df = sfl.read_file(sfl_file)
    df = sfl.fix(df)
    dup_files, df = sfl.dedup(df)
    if len(dup_files):
        click.echo(os.linesep.join(['{}\t{}'.format(*d) for d in dup_files]), err=True)
    sfl.save_to_file(df, sys.stdout)


@sfl_cmd.command('fix-event-rate')
@click.argument('sfl-file', nargs=1, type=click.Path(exists=True, readable=True))
@click.argument('events-file', nargs=1, type=click.Path(exists=True, readable=True))
def sfl_fix_event_rate_cmd(sfl_file, events_file):
    """
    Calculates true event rates.

    EVENTS-FILE should be a TSV file with EVT path/file ID in first
    column and event count in last column, or a popcycle SQLite3 database file
    with a '.db' extension. A version of SFL_FILE with updated event rates will
    be printed to STDOUT. In cases where the file duration value is < 0 or NA
    the event rate will be NA.
    """
    df = sfl.read_file(sfl_file)
    df = sfl.fix(df)

    # Event counts should be a dict of { file: event_count }
    if events_file.endswith(".db"):
        event_counts = db.get_event_counts(events_file)
    else:
        lines = [x.rstrip().split('\t') for x in events_file.readlines()]
        event_counts = {seaflowfile.SeaFlowFile(x[0]).file_id: int(x[-1]) for x in lines}

    df = sfl.fix_event_rate(df, event_counts)
    sfl.save_to_file(df, sys.stdout)



@sfl_cmd.command('fix-underway')
@click.argument('sfl-file', nargs=1, type=click.Path(exists=True, readable=True))
@click.argument('cruisemic-dir', nargs=1, type=click.Path(exists=True, readable=True))
def sfl_fix_underway_cmd(sfl_file, cruisemic_dir):
    """
    Replace SFL underway data with data from cruisemic.

    A version of SFL_FILE with updated underway columns based on data from the
    cruisemic output directory CRUISEMIC_DIR will be printed to STDOUT.
    """
    df_sfl = sfl.read_file(sfl_file)
    df_sfl = sfl.fix(df_sfl)

    # Read cruisemic metadata file to find underway data
    meta_glob = glob(os.path.join(cruisemic_dir, "metadata*"))
    if len(meta_glob) != 1:
        raise click.ClickException("could not find geo-file")
    with fileio.file_open_r(meta_glob[0], as_text=True) as metafh:
        meta = json.load(metafh)

    geo_filename = meta['GeoFeed']
    thermo_filename = meta['ThermoFeed']
    lat_col = meta['LatitudeCol']
    lon_col = meta['LongitudeCol']
    temp_col = meta['TemperatureCol']
    sal_col = meta['SalinityCol']
    cond_col = meta['ConductivityCol']

    # Read underway data
    geo_glob = glob(os.path.join(cruisemic_dir, geo_filename) + "*")
    thermo_glob = glob(os.path.join(cruisemic_dir, thermo_filename) + "*")
    if len(geo_glob) != 1:
        raise click.ClickException("could not find cruisemic geo data file")
    if len(thermo_glob) != 1:
        raise click.ClickException("could not find cruisemic thermosalinograph file")
    with fileio.file_open_r(geo_glob[0], as_text=True) as geo_fh:
        geo_df = tsdataformat.read_tsdata(geo_fh, convert="time")
        geo_df["lat"] = geo_df[lat_col].astype(float)
        geo_df["lon"] = geo_df[lon_col].astype(float)
        geo_df = geo_df[["time", "lat", "lon"]]
    with fileio.file_open_r(thermo_glob[0], as_text=True) as thermo_fh:
        thermo_df = tsdataformat.read_tsdata(thermo_fh, convert="time")
        thermo_df["ocean_tmp"] = thermo_df[temp_col].astype(float)
        thermo_df["salinity"] = thermo_df[sal_col].astype(float)
        thermo_df["conductivity"] = thermo_df[cond_col].astype(float)
        thermo_df = thermo_df[["time", "ocean_tmp", "salinity", "conductivity"]]

    df = sfl.fix_underway(df_sfl, geo_df, thermo_df)
    sfl.save_to_file(df, sys.stdout)


@sfl_cmd.command('manifest')
@click.option('-v', '--verbose', is_flag=True,
    help='Print a list of all file ids not in common between SFL and directory.')
@click.argument('sfl-file', nargs=1, type=click.File())
@click.argument('evt-dir', nargs=1, type=str)
def manifest_cmd(verbose, sfl_file, evt_dir):
    """
    Compares files in SFL-FILE with files in EVT-DIR.

    If EVT-DIR begins with 's3://s3-bucket-name' then files will located in S3.
    To configure credentials for S3 access use the 'aws' command-line tool from
    the 'awscli' Python package.

    It's normal for about one file per day to be missing from the SFL file or
    EVT day of year folder, especially around midnight.

    To read from STDIN use '-' for SFL_FILE. Prints a file list diff to STDOUT.
    """
    found_evt_ids = []
    if evt_dir.startswith("s3://"):
        try:
            _, _, bucket, evt_dir = evt_dir.split("/", 3)
        except ValueError:
            raise click.ClickException("could not parse bucket and folder from S3 EVT-DIR")
        try:
            files = cloud.get_s3_file_list(bucket, evt_dir)
        except botocore.exceptions.NoCredentialsError:
            print('Please configure aws first:', file=sys.stderr)
            print('  $ pip install awscli', file=sys.stderr)
            print('  then', file=sys.stderr)
            print('  $ aws configure', file=sys.stderr)
            raise click.Abort()
        except sfperrors.S3Error as e:
            raise click.ClickException(e)
        found_evt_files = seaflowfile.sorted_files(seaflowfile.keep_evt_files(files))
    else:
        try:
            found_evt_files = seaflowfile.find_evt_files(evt_dir)
        except OSError as e:
            raise click.ClickException(e)

    df = sfl.read_file(sfl_file)
    sfl_evt_ids = [seaflowfile.SeaFlowFile(f).file_id for f in df['file']]
    found_evt_ids = [seaflowfile.SeaFlowFile(f).path_file_id for f in found_evt_files]
    sfl_set = set(sfl_evt_ids)
    found_set = set(found_evt_ids)

    print('%d EVT files in SFL file %s' % (len(sfl_set), sfl_file.name))
    print('%d EVT files in directory %s' % (len(found_set), evt_dir))
    print('%d EVT files in common' % len(sfl_set.intersection(found_set)))
    if verbose and \
       (len(sfl_set.intersection(found_set)) != len(sfl_set) or
        len(sfl_set.intersection(found_set)) != len(found_set)):
        print('')
        print('EVT files in SFL but not found:')
        print(os.linesep.join(sorted(sfl_set.difference(found_set))))
        print('')
        print('EVT files found but not in SFL:')
        print(os.linesep.join(sorted(found_set.difference(sfl_set))))
        print('')


@sfl_cmd.command('print')
@click.argument('sfl-files', metavar='SFL', nargs=-1, type=click.Path(exists=True))
def sfl_print_cmd(sfl_files):
    """
    Concatenates raw SFL files, prints a standardized SFL file.

    Makes the following changes to create a standardized SFL file:

    - Outputs only columns for database import.

    - The correct day of year folder will be added to FILE column values if not
    present

    - DATE column will be created if not present based on "FILE" column values
    (only applies to new-style datestamped file names)

    - STREAM PRESSURE values <= 0 will be changed to 1e-4

    - Any other required columns which are missing will be created with NA
    values.

    Input files will be concatenated in the order they're listed on the
    command-line. Outputs to STDOUT.
    """
    dfs = []
    for f in sfl_files:
        onedf = sfl.read_file(f)
        onedf = sfl.fix(onedf)
        dfs.append(onedf)
    df = pd.concat(dfs, ignore_index=True)
    sfl.save_to_file(df, sys.stdout)


@sfl_cmd.command('validate')
@click.option('-a', '--all', 'report_all', is_flag=True,
    help='Report all errors.')
@click.argument('sfl-file', nargs=-1, type=click.Path(exists=True, readable=True))
def sfl_validate_cmd(report_all, sfl_file):
    """
    Validates SFL files.

    Checks that:

    - Required columns are present: FILE, DATE, FILE DURATION, LAT, LON,
    CONDUCTIVITY, SALINITY, OCEAN TEMP, PAR, BULK RED, STREAM PRESSURE,
    EVENT RATE

    - No missing values in following columns: FILE, DATE, FILE DURATION, LAT,
    LON, STREAM PRESSURE, EVENT RATE

    - FILE column values have day of year folders, are in the proper format, in
    chronological order, are unique, and matches DATE column

    - DATE column values are in the proper format, represent valid dates and
    times, and are UTC

    - FILE DURATION is a positive number

    - LAT and LON column values are valid decimal degree values in the correct
    ranges

    - CONDUCTIVITY, SALINITY, OCEAN TEMP, PAR, BULK_RED column values are
    numbers

    - STREAM PRESSURE is a positive number >= 1e-4

    - EVENT RATE is a positive number

    Because some of these errors can affect every row of the file (e.g. out of
    order files), only the first error of each type is printed. To get a full
    printout of all errors use --all.

    Prints error report to STDOUT.
    """
    need_header = True
    for f in sfl_file:
        try:
            df = sfl.read_file(f)
        except sfperrors.FileError as e:
            raise click.ClickException(str(e))
        errors = sfl.check(df)
        if len(errors) > 0:
            sfl.print_tsv_errors(errors, sys.stdout, os.path.basename(f),
                                 print_all=report_all, header=need_header)
        need_header = False



import itertools
import multiprocessing as mp
import random

import pandas as pd
from seaflowpy import fileio
from seaflowpy import particleops
from seaflowpy import seaflowfile
from seaflowpy import util


def random_select(things, fraction, seed=None):
    """
    Randomly sample a fraction of items in an iterable.

    Parameters
    ----------
    things: list
        Things to sample from.
    fraction: float
        Fraction of things to sample, between 0 and 1. Things will be randomly
        chosen. At least 1 thing will always be chosen.
    seed: int
        Integer seed for PRNG. If None, a source of random seed will be used.

    Returns
    -------
    list
        Chosen things.
    """
    if (seed is not None) and not isinstance(seed, int):
        raise ValueError("seed must be an None or an int")
    if fraction < 0 or fraction > 1:
        raise ValueError("fraction must be between 0 and 1")

    if len(things) == 0:
        return things

    if seed is not None:
        rand = random.Random(seed)
    else:
        rand = random.Random()
    count = max(int(len(things) * fraction), 1)  # select at least 1 thing
    thing_indexes = list(range(0, len(things)))
    chosen_i = sorted(rand.sample(thing_indexes, count))
    chosen_things = [things[i] for i in chosen_i]
    return chosen_things


def sample(
    evtpaths,
    n,
    outpath,
    dates=None,
    min_chl=0,
    min_fsc=0,
    min_pe=0,
    multi=False,
    noise_filter=False,
    process_count=1,
    seed=None,
):
    """
    Randomly sample rows from EVT files.

    May sample many files separately or as one data set. This function is a
    parallel wrapper for sample_many_to_one.

    Parameters
    ----------
    evtpaths: list of str
        EVT file paths.
    n: int
        Events to keep per file (--multi) or total.
    outpath: str
        Parquet output file.
    dates: dict of {file_id : datetime.datetime}
        If provided, create a column of dates called "date".
    min_chl: int, default 0
        Minimum chl_small value.
    min_fsc: int, default 0
        Minimum fsc_small value.
    min_pe: int, default 0
        Minimum pe value.
    multi: bool, default False
        Output one file per input file. If False, create one aggregated output file.
    noise_filter: bool, default False
        Remove noise particles after sampling.
    process_count: int, default: 1
        Number of worker processes to create.
    seed: int, default None
        Integer seed for PRNG, used in sampling files and events. If None, a
        source of random seed will be used.

    Returns
    -------
    tupe of (list of dicts for each file, unhandled exceptions):
        [
            [{
                "msg": any errors encountered while processing this file,
                "file_id": input EVT file ID,
                "events": event count in original file,
                "events_postfilter": event count after applying min val / noise filters,
                "events_postsampling": event count after subsampling
            }, ...],
            [unhandled exceptions]
    """
    if len(evtpaths) == 0:
        return ([], [])

    # Don't create more processes than input files
    process_count = min(process_count, len(evtpaths))
    # Split EVT files up into buckets, one for each worker process
    file_buckets = util.jobs_parts(evtpaths, process_count)

    # How many events to take per file
    if multi:
        n_per_file = n
    else:
        n_per_file = n // len(evtpaths)

    pool = mp.Pool(processes=process_count)

    # Result handling callbacks for mp.async_apply
    mp_results, mp_errs = [], []

    def cb(result):
        mp_results.append(result)

    def err_cb(err):
        mp_errs.append(err)

    # kwargs for each worker process, same for each
    kwargs = {
        "min_chl": min_chl,
        "min_fsc": min_fsc,
        "min_pe": min_pe,
        "noise_filter": noise_filter,
        "seed": seed,
    }

    for i, bucket_o_files in enumerate(file_buckets):
        args = (i, bucket_o_files, n_per_file)
        pool.apply_async(
            _sample_many_to_one_worker,
            args,
            kwargs,
            callback=cb,
            error_callback=err_cb,
        )
    pool.close()
    pool.join()
    mp_results = sorted(
        mp_results, key=lambda x: x["i"]
    )  # sort async results by orig order
    results = list(
        itertools.chain.from_iterable([r["results"] for r in mp_results])
    )
    df = pd.concat([r["df"] for r in mp_results], ignore_index=True)
    if dates:
        df["date"] = df["file_id"].map(dates)
    df["file_id"] = df["file_id"].astype("category")
    assert len(df.index) == sum([r["events_postsampling"] for r in results])
    df.to_parquet(outpath)

    return (results, mp_errs)


def _sample_many_to_one_worker(i, *args, **kwargs):
    results = sample_many_to_one(*args, **kwargs)
    results["i"] = i  # to sort async result blocks later
    return results


def sample_many_to_one(
    evtpaths, n, min_chl=0, min_fsc=0, min_pe=0, noise_filter=False, seed=None,
):
    """
    Randomly sample rows from EVT files and combined into one dataframe.

    Parameters
    ----------
    evtpaths: list of str
        EVT file paths.
    n: int
        Events to sample from each input file, > 0.
    min_chl: int, default 0
        Minimum chl_small value.
    min_fsc: int, default 0
        Minimum fsc_small value.
    min_pe: int, default 0
        Minimum pe value.
    noise_filter: bool, default False
        Remove noise particles after sampling.
    seed: int, default None
        Integer seed for PRNG, used in sampling files and events. If None, a
        source of random seed will be used.

    Returns
    -------
    dict
        {
            "df": combined dataframe of sampled data from all files,
            "results": [{
                "msg": any errors encountered while processing this file,
                "file_id": input EVT file ID,
                "events": event count in original file,
                "events_postfilter": event count after applying min val / noise filters,
                "events_postsampling": event count after subsampling
            }, ...]
        }
    """
    columns = ["D1", "D2", "fsc_small", "pe", "chl_small"]
    results = []

    for f in evtpaths:
        msg = ""
        try:
            data = fileio.read_evt_labview(f)
            df = data["df"]
        except Exception as e:
            msg = "{}: {}".format(type(e).__name__, str(e))
            df = particleops.empty_df()
        result = sample_one(
            df,
            n,
            min_chl=min_chl,
            min_fsc=min_fsc,
            min_pe=min_pe,
            noise_filter=noise_filter,
            seed=seed,
        )
        result["df"] = result["df"][columns]
        file_id = seaflowfile.SeaFlowFile(f).file_id
        result["df"]["file_id"] = file_id
        result["file_id"] = file_id
        result["msg"] = msg
        results.append(result)

    if len(results):
        df = pd.concat([r["df"] for r in results], ignore_index=True)
        for r in results:
            del r["df"]
    else:
        df = particleops.empty_df()[columns]
        df["file_id"] = None

    return {
        "df": df,
        "results": results,
    }


def sample_one(df, n, noise_filter=True, min_chl=0, min_fsc=0, min_pe=0, seed=None):
    """
    Randomly sample rows from an EVT dataframe.

    Parameters
    ----------
    df: pandas.DataFrame
        EVT dataframe to sample from.
    n: int
        Events to sample from input files, > 0.
    noise_filter: bool, default True
        Remove noise particles after sampling.
    min_chl: int, default 0
        Minimum chl_small value.
    min_fsc: int, default 0
        Minimum fsc_small value.
    min_pe: int, default 0
        Minimum pe value.
    seed: int, default None
        Integer seed for PRNG, used in sampling files and events. If None, a
        source of random seed will be used.

    Raises
    ------
    ValueError

    Returns
    -------
    dict
        {
            "df": subsampled pandas.DataFrame,
            "events": event count in original file,
            "events_postfilter": event count after applying min val / noise filters,
            "events_postsampling": event count after subsampling
        }
    """
    if (seed is not None) and not isinstance(seed, int):
        raise ValueError("seed must be an None or an int")
    if n <= 0:
        raise ValueError("n must be > 0")

    events = len(df.index)
    chl = df["chl_small"].values >= min_chl
    fsc = df["fsc_small"].values >= min_fsc
    pe = df["pe"].values >= min_pe
    if noise_filter:
        noise = particleops.mark_noise(df)
        df = df[(~noise) & chl & fsc & pe]
    else:
        df = df[chl & fsc & pe]
    events_postfilter = len(df.index)
    try:
        frac = min(n / events_postfilter, 1)
    except ZeroDivisionError:
        pass
    else:
        if seed is None:
            df = df.sample(frac=frac)
        else:
            df = df.sample(frac=frac, random_state=seed)

    return {
        "df": df,
        "events": events,
        "events_postfilter": events_postfilter,
        "events_postsampling": len(df.index),
    }



import boto3
import botocore

from . import errors

def get_s3_file_list(bucket_name, prefix):
    """Get list of S3 object keys for one folder"""
    while prefix.endswith("/"):
        prefix = prefix[:-1]
    prefix = prefix + "/"
    s3 = boto3.resource("s3")
    bucket = s3.Bucket(bucket_name)
    exists = True
    try:
        s3.meta.client.head_bucket(Bucket=bucket_name)
    except botocore.exceptions.ClientError as e:
        # If a client error is thrown, then check that it was a 404 error.
        # If it was a 404 error, then the bucket does not exist.
        error_code = int(e.response['Error']['Code'])
        if error_code == 404:
            exists = False
    if not exists:
        raise errors.S3Error(f"S3 bucket {bucket_name} does not exist")

    files = []
    for obj in bucket.objects.filter(Prefix=prefix):
        files.append(obj.key)
    return files



from setuptools import setup, find_packages
import versioneer

setup(
    name='seaflowpy',
    description='A Python library for SeaFlow data.',
    long_description=open('README.md', 'r').read(),
    long_description_content_type='text/markdown',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    url='https://github.com/armbrustlab/seaflowpy',
    author='Chris T. Berthiaume',
    author_email='chrisbee@uw.edu',
    packages=find_packages(where='src'),
    package_dir={'': 'src'},
    include_package_data=True,
    python_requires='>=3.7.1',
    install_requires=[
        'boto3',
        'click',
        'joblib',
        'pandas',
        'pyarrow',
        'tsdataformat'
    ],
    entry_points={
        'console_scripts': [
            'seaflowpy=seaflowpy.cli.cli:cli'
        ]
    },
    classifiers=[
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)'
    ],
    zip_safe=True
)



from . import cloud
from . import db
from . import errors
from . import fileio
from . import filterevt
from . import geo
from . import particleops
from . import sample
from . import sfl
from . import time
from . import util

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



import datetime

import pandas as pd
import pytest
import seaflowpy as sfp

# pylint: disable=redefined-outer-name


def test_invalid_filename():
    with pytest.raises(sfp.errors.FileError):
        _ = sfp.seaflowfile.SeaFlowFile("foobar")
    with pytest.raises(sfp.errors.FileError):
        _ = sfp.seaflowfile.SeaFlowFile("foo/bar/2014-07-04T00-00-02+00-00.opp.gz")
    with pytest.raises(sfp.errors.FileError):
        _ = sfp.seaflowfile.SeaFlowFile("foo/bar/2014-07-04T00-00-02+00-00.vct.gz")
    with pytest.raises(sfp.errors.FileError):
        _ = sfp.seaflowfile.SeaFlowFile("foo/bar/2014-07-04T00-00-02+00-00.sfl")

def test_invalid_filename_date():
    with pytest.raises(sfp.errors.FileError):
        _ = sfp.seaflowfile.SeaFlowFile("2014-07-32T00-00-02+00-00")

def test_new_style():
    f = sfp.seaflowfile.SeaFlowFile("2014-07-04T00-00-02+00-00")
    assert f.path == "2014-07-04T00-00-02+00-00"
    assert f.filename == "2014-07-04T00-00-02+00-00"
    assert f.filename_nogz == "2014-07-04T00-00-02+00-00"
    assert f.file_id == "2014_185/2014-07-04T00-00-02+00-00"
    assert f.path_file_id == "2014-07-04T00-00-02+00-00"
    assert f.dayofyear == "2014_185"
    assert f.path_dayofyear == ''
    assert f.is_old_style is False
    assert f.is_new_style is True

    f = sfp.seaflowfile.SeaFlowFile("2014_185/2014-07-04T00-00-02+00-00")
    assert f.path == "2014_185/2014-07-04T00-00-02+00-00"
    assert f.filename == "2014-07-04T00-00-02+00-00"
    assert f.filename_nogz == "2014-07-04T00-00-02+00-00"
    assert f.file_id == "2014_185/2014-07-04T00-00-02+00-00"
    assert f.path_file_id == "2014_185/2014-07-04T00-00-02+00-00"
    assert f.dayofyear == "2014_185"
    assert f.path_dayofyear == "2014_185"
    assert f.is_old_style is False
    assert f.is_new_style is True

    f = sfp.seaflowfile.SeaFlowFile("foo/2014-07-04T00-00-02+00-00")
    assert f.path == "foo/2014-07-04T00-00-02+00-00"
    assert f.filename == "2014-07-04T00-00-02+00-00"
    assert f.filename_nogz == "2014-07-04T00-00-02+00-00"
    assert f.file_id == "2014_185/2014-07-04T00-00-02+00-00"
    assert f.path_file_id == "2014-07-04T00-00-02+00-00"
    assert f.dayofyear == "2014_185"
    assert f.path_dayofyear == ''
    assert f.is_old_style is False
    assert f.is_new_style is True

    f = sfp.seaflowfile.SeaFlowFile("foo/2014_185/2014-07-04T00-00-02+00-00")
    assert f.path == "foo/2014_185/2014-07-04T00-00-02+00-00"
    assert f.filename == "2014-07-04T00-00-02+00-00"
    assert f.filename_nogz == "2014-07-04T00-00-02+00-00"
    assert f.file_id == "2014_185/2014-07-04T00-00-02+00-00"
    assert f.path_file_id == "2014_185/2014-07-04T00-00-02+00-00"
    assert f.dayofyear == "2014_185"
    assert f.path_dayofyear == "2014_185"
    assert f.is_old_style is False
    assert f.is_new_style is True

    f = sfp.seaflowfile.SeaFlowFile("foo/bar/2014-07-04T00-00-02+00-00")
    assert f.path == "foo/bar/2014-07-04T00-00-02+00-00"
    assert f.filename == "2014-07-04T00-00-02+00-00"
    assert f.filename_nogz == "2014-07-04T00-00-02+00-00"
    assert f.file_id == "2014_185/2014-07-04T00-00-02+00-00"
    assert f.path_file_id == "2014-07-04T00-00-02+00-00"
    assert f.dayofyear == "2014_185"
    assert f.path_dayofyear == ''
    assert f.is_old_style is False
    assert f.is_new_style is True

    f = sfp.seaflowfile.SeaFlowFile("foo/bar/2014-07-04T00-00-02+00-00.gz")
    assert f.path == "foo/bar/2014-07-04T00-00-02+00-00.gz"
    assert f.filename == "2014-07-04T00-00-02+00-00.gz"
    assert f.filename_nogz == "2014-07-04T00-00-02+00-00"
    assert f.file_id == "2014_185/2014-07-04T00-00-02+00-00"
    assert f.path_file_id == "2014-07-04T00-00-02+00-00"
    assert f.dayofyear == "2014_185"
    assert f.path_dayofyear == ''
    assert f.is_old_style is False
    assert f.is_new_style is True


def test_old_style():
    f = sfp.seaflowfile.SeaFlowFile("42.evt")
    assert f.path == "42.evt"
    assert f.filename == "42.evt"
    assert f.filename_nogz == "42.evt"
    assert f.file_id == "42.evt"
    assert f.path_file_id == f.file_id
    assert f.dayofyear == ''
    assert f.path_dayofyear == ''
    assert f.is_old_style is True
    assert f.is_new_style is False

    f = sfp.seaflowfile.SeaFlowFile("2014_185/42.evt")
    assert f.path == "2014_185/42.evt"
    assert f.filename == "42.evt"
    assert f.filename_nogz == "42.evt"
    assert f.file_id == "2014_185/42.evt"
    assert f.path_file_id == f.file_id
    assert f.dayofyear == ''
    assert f.path_dayofyear == "2014_185"
    assert f.is_old_style is True
    assert f.is_new_style is False

    f = sfp.seaflowfile.SeaFlowFile("foo/42.evt")
    assert f.path == "foo/42.evt"
    assert f.filename == "42.evt"
    assert f.filename_nogz == "42.evt"
    assert f.path_file_id == f.file_id
    assert f.file_id == "42.evt"
    assert f.dayofyear == ''
    assert f.path_dayofyear == ''
    assert f.is_old_style is True
    assert f.is_new_style is False

    f = sfp.seaflowfile.SeaFlowFile("foo/2014_185/42.evt")
    assert f.path == "foo/2014_185/42.evt"
    assert f.filename == "42.evt"
    assert f.filename_nogz == "42.evt"
    assert f.file_id == "2014_185/42.evt"
    assert f.path_file_id == f.file_id
    assert f.dayofyear == ''
    assert f.path_dayofyear == "2014_185"
    assert f.is_old_style is True
    assert f.is_new_style is False

    f = sfp.seaflowfile.SeaFlowFile("foo/bar/42.evt")
    assert f.path == "foo/bar/42.evt"
    assert f.filename == "42.evt"
    assert f.filename_nogz == "42.evt"
    assert f.file_id == "42.evt"
    assert f.path_file_id == f.file_id
    assert f.dayofyear == ''
    assert f.path_dayofyear == ''
    assert f.is_old_style is True
    assert f.is_new_style is False

    f = sfp.seaflowfile.SeaFlowFile("foo/bar/42.evt.gz")
    assert f.path == "foo/bar/42.evt.gz"
    assert f.filename == "42.evt.gz"
    assert f.filename_nogz == "42.evt"
    assert f.file_id == "42.evt"
    assert f.path_file_id == f.file_id
    assert f.dayofyear == ''
    assert f.path_dayofyear == ''
    assert f.is_old_style is True
    assert f.is_new_style is False


def test_set_date():
    # Should do nothing for new style if matches
    date = datetime.datetime.fromisoformat("2014-07-04T00:00:02+00:00")
    f = sfp.seaflowfile.SeaFlowFile("2014_185/2014-07-04T00-00-02+00-00", date=date)
    assert f.rfc3339 == "2014-07-04T00:00:02+00:00"

    # Should raise ValueError for new style if doesn't match
    date = datetime.datetime.fromisoformat("2014-07-05T00:00:02+00:00")
    with pytest.raises(ValueError):
        _ = sfp.seaflowfile.SeaFlowFile("2014_185/2014-07-04T00-00-02+00-00", date=date)

    # Should set data for old style
    date = datetime.datetime.fromisoformat("2014-07-04T02:06:00+00:00")
    f = sfp.seaflowfile.SeaFlowFile("2014_185/42.evt", date=date)
    assert f.file_id == "2014_185/42.evt"
    assert f.dayofyear == "2014_185"
    assert f.rfc3339 == "2014-07-04T02:06:00+00:00"


def test_sort_new_chronologically():
    unsorted_files = [
        "2014_186/2014-07-06T00-00-05+00-00",
        "2014_185/2014-07-04T10-00-02+00-00",
        "2014_185/2014-07-04T01-00-02+00-00",
        "2014_186/2014-07-05T00-00-02+00-00"
    ]
    sorted_files = [
        "2014_185/2014-07-04T01-00-02+00-00",
        "2014_185/2014-07-04T10-00-02+00-00",
        "2014_186/2014-07-05T00-00-02+00-00",
        "2014_186/2014-07-06T00-00-05+00-00"
    ]

    assert sfp.seaflowfile.sorted_files(unsorted_files) == sorted_files


def test_sorted_files():
    unsorted_files = [
        "2014_186/100.evt",
        "2014-12-08T22-56-34+00-00",
        "2014_185/10.evt",
        "9.evt",
        "2014_342/2014-12-08T22-53-34+00-00",
        "2014_350/2014-12-08T22-51-34+00-00",
        "2014_185/1.evt",
        "2014_186/23.evt",
        "2015_186/1.evt"
    ]
    sorted_files = [
        "9.evt",
        "2014_185/1.evt",
        "2014_185/10.evt",
        "2014_186/23.evt",
        "2014_186/100.evt",
        "2014_350/2014-12-08T22-51-34+00-00",  # note dayofyear_path from path is wrong
        "2014_342/2014-12-08T22-53-34+00-00",
        "2014-12-08T22-56-34+00-00",
        "2015_186/1.evt"
    ]

    assert sfp.seaflowfile.sorted_files(unsorted_files) == sorted_files


def test_date_from_file_name():
    files = [
        "2014-07-06T00-00-05+00-00",
        "2014-07-06T00-00-05-00-00",
        "2014-07-06T00-00-05-07-00",  # all dates assumed UTC, even if non-UTC TZ offset
        "2014-07-06T00-00-05+00-00.gz"
    ]
    answers = [
        "2014-07-06T00:00:05+00:00",
        "2014-07-06T00:00:05+00:00",
        "2014-07-06T00:00:05+00:00",  # all dates assumed UTC, even if non-UTC TZ offset
        "2014-07-06T00:00:05+00:00"
    ]
    for i, f in enumerate(files):
        s = sfp.seaflowfile.SeaFlowFile(f)
        assert s.rfc3339 == answers[i]


def test_dayofyear_from_filename():
    files = [
        "2014-07-06T00-00-05+00-00",
        "2014-07-06T00-00-05-00-00",
        "2014-07-06T00-00-05-07-00",
        "2014-07-06T00-00-05+00-00.gz",
        "2014_001/2014-07-06T00-00-05+00-00",
        "2014_001/2014-07-06T00-00-05-00-00",
        "2014_001/2014-07-06T00-00-05-07-00",
        "2014_001/2014-07-06T00-00-05+00-00.gz"
    ]
    dayofyear_answer = "2014_187"
    path_dayofyear_answer = "2014_001"

    for f in files[0:4]:
        s = sfp.seaflowfile.SeaFlowFile(f)
        assert s.dayofyear == dayofyear_answer
        assert s.path_dayofyear == ''

    for f in files[4:]:
        s = sfp.seaflowfile.SeaFlowFile(f)
        assert s.dayofyear == dayofyear_answer
        assert s.path_dayofyear == path_dayofyear_answer

def test_parse_path():
    files = [
        "2014_187/2014-07-06T00-00-05-00-00",
        "2014-07-06T00-00-05-00-00",
        "2014_187/2014-07-06T00-00-05-00-00.gz",
        "2014-07-06T00-00-05-00-00.gz"
    ]

    answers = [
        {"file": "2014-07-06T00-00-05-00-00", "dayofyear": "2014_187"},
        {"file": "2014-07-06T00-00-05-00-00", "dayofyear": ''},
        {"file": "2014-07-06T00-00-05-00-00.gz", "dayofyear": "2014_187"},
        {"file": "2014-07-06T00-00-05-00-00.gz", "dayofyear": ''}
    ]

    for i, f in enumerate(files):
        assert sfp.seaflowfile.parse_path(f) == answers[i]

def test_fix_file_id():
    origid = "2014_001/2014-12-08T22-53-34+00-00"
    correctid = "2014_342/2014-12-08T22-53-34+00-00"
    assert sfp.seaflowfile.SeaFlowFile(origid).file_id == correctid
    assert sfp.seaflowfile.SeaFlowFile(origid).path_file_id == origid


def test_file_list_filter():
    files = [
        "tests/testcruise_evt/2014_185/2014-07-04T00-00-02+00-00",
        "tests/testcruise_evt/2014_185/2014-07-04T00-03-02+00-00.gz",
        "tests/testcruise_evt/2014_185/2014-07-04T00-06-02+00-00",
        "tests/testcruise_evt/2014_185/2014-07-04T00-09-02+00-00",
        "tests/testcruise_evt/2014_185/2014-07-04T00-12-02+00-00",
        "tests/testcruise_evt/2014_185/2014-07-04T00-15-02+00-00.gz",
        "tests/testcruise_evt/2014_185/2014-07-04T00-17-02+00-00.gz",
        "tests/testcruise_evt/2014_185/2014-07-04T00-21-02+00-00"
    ]
    filter_list = [
        "2014_185/2014-07-04T00-00-02+00-00",
        "testcruise_evt/2014_185/2014-07-04T00-03-02+00-00.gz",
        "foo/2014_185/2014-07-04T00-06-02+00-00",
        "2014_185/2014-07-04T00-17-02+00-00.gz",
        "2014_185/2014-07-04T00-30-02+00-00.gz"
    ]
    answer = [
        "tests/testcruise_evt/2014_185/2014-07-04T00-00-02+00-00",
        "tests/testcruise_evt/2014_185/2014-07-04T00-03-02+00-00.gz",
        "tests/testcruise_evt/2014_185/2014-07-04T00-06-02+00-00",
        "tests/testcruise_evt/2014_185/2014-07-04T00-17-02+00-00.gz"
    ]
    result = sfp.seaflowfile.filtered_file_list(files, filter_list)
    assert answer == result

def test_keep_evt_files():
    files = [
        "testcruise/2014_185/100.evt",
        "testcruise/2014_185/200.evt.gz",
        "not_evt_file",
        "testcruise/2014_185/2014-07-04T00-00-02+00-00",
        "testcruise/2014_185/2014-07-04T00-03-02+00-00.gz",
    ]
    parsed = sfp.seaflowfile.keep_evt_files(files)
    assert parsed == (files[:2] + files[3:])

def test_find_evt_files():
    files = sfp.seaflowfile.find_evt_files("tests/testcruise_evt")
    answer = [
        "tests/testcruise_evt/2014_185/2014-07-04T00-00-02+00-00",
        "tests/testcruise_evt/2014_185/2014-07-04T00-03-02+00-00.gz",
        "tests/testcruise_evt/2014_185/2014-07-04T00-06-02+00-00",
        "tests/testcruise_evt/2014_185/2014-07-04T00-09-02+00-00",
        "tests/testcruise_evt/2014_185/2014-07-04T00-12-02+00-00",
        "tests/testcruise_evt/2014_185/2014-07-04T01-15-02+00-00.gz",
        "tests/testcruise_evt/2014_185/2014-07-04T01-17-02+00-00.gz",
        "tests/testcruise_evt/2014_185/2014-07-04T01-21-02+00-00",
        "tests/testcruise_evt/2014_185/2014-07-04T01-27-02+00-00",
        "tests/testcruise_evt/2014_185/2014-07-04T01-30-02+00-00"
    ]
    assert files == answer

def test_timeselect_evt_files():
    raw_files = [
        "tests/testcruise_evt/2014_185/2014-07-04T00-00-02+00-00",
        "tests/testcruise_evt/2014_185/2014-07-04T00-03-02+00-00.gz",
        "tests/testcruise_evt/2014_185/2014-07-04T00-06-02+00-00",
        "tests/testcruise_evt/2014_185/2014-07-04T00-09-02+00-00"
    ]
    files = [sfp.seaflowfile.SeaFlowFile(f) for f in raw_files]

    # Exact boundaries, test inclusivity
    time_start = datetime.datetime.fromisoformat("2014-07-04T00:03:02+00:00")
    time_end = datetime.datetime.fromisoformat("2014-07-04T00:06:02+00:00")
    selected = sfp.seaflowfile.timeselect_evt_files(files, time_start, time_end)
    answer = [
        "tests/testcruise_evt/2014_185/2014-07-04T00-03-02+00-00.gz",
        "tests/testcruise_evt/2014_185/2014-07-04T00-06-02+00-00"
    ]
    selected = [f.path for f in selected]
    assert selected == answer

    # Inexact boundaries
    time_start = datetime.datetime.fromisoformat("2014-07-04T00:03:00+00:00")
    time_end = datetime.datetime.fromisoformat("2014-07-04T00:07:00+00:00")
    selected = sfp.seaflowfile.timeselect_evt_files(files, time_start, time_end)
    answer = [
        "tests/testcruise_evt/2014_185/2014-07-04T00-03-02+00-00.gz",
        "tests/testcruise_evt/2014_185/2014-07-04T00-06-02+00-00"
    ]
    selected = [f.path for f in selected]
    assert selected == answer

    # No start
    time_end = datetime.datetime.fromisoformat("2014-07-04T00:06:00+00:00")
    selected = sfp.seaflowfile.timeselect_evt_files(files, None, time_end)
    answer = [
        "tests/testcruise_evt/2014_185/2014-07-04T00-00-02+00-00",
        "tests/testcruise_evt/2014_185/2014-07-04T00-03-02+00-00.gz"
    ]
    selected = [f.path for f in selected]
    assert selected == answer

    # No end
    time_start = datetime.datetime.fromisoformat("2014-07-04T00:06:00+00:00")
    selected = sfp.seaflowfile.timeselect_evt_files(files, time_start, None)
    answer = [
        "tests/testcruise_evt/2014_185/2014-07-04T00-06-02+00-00",
        "tests/testcruise_evt/2014_185/2014-07-04T00-09-02+00-00"
    ]
    selected = [f.path for f in selected]
    assert selected == answer

    # Nothing is selected
    time_start = datetime.datetime.fromisoformat("2014-07-01T00:03:02+00:00")
    time_end = datetime.datetime.fromisoformat("2014-07-03T00:06:02+00:00")
    selected = sfp.seaflowfile.timeselect_evt_files(files, time_start, time_end)
    assert selected == []

    # No dates
    selected = sfp.seaflowfile.timeselect_evt_files(files, None, None)
    selected = [f.path for f in selected]
    assert selected == raw_files

    # No tz offset
    time_start = datetime.datetime.fromisoformat("2014-07-04T00:03:02")
    time_end = datetime.datetime.fromisoformat("2014-07-04T00:06:02")
    with pytest.raises(TypeError):
        selected = sfp.seaflowfile.timeselect_evt_files(files, time_start, time_end)

    # Set date to None for one of the files
    files_tmp = [sfp.seaflowfile.SeaFlowFile(f) for f in raw_files]
    files_tmp[2].date = None  # erase date
    time_start = datetime.datetime.fromisoformat("2014-07-04T00:06:00+00:00")
    selected = sfp.seaflowfile.timeselect_evt_files(files_tmp, time_start, None)
    answer = ["tests/testcruise_evt/2014_185/2014-07-04T00-09-02+00-00"]
    selected = [f.path for f in selected]
    assert selected == answer


def test_date_evt_files():
    evt_files = [
        "tests/testcruise_evt/2014_185/2014-07-04T00-00-02+00-00",
        "tests/testcruise_evt/2014_185/2014-07-04T00-03-02+00-00.gz",
        "tests/testcruise_evt/2014_185/2014-07-04T00-06-02+00-00",
        "tests/testcruise_evt/2014_185/2014-07-04T00-09-02+00-00",
        "tests/testcruise_evt/2014_185/2014-07-04T00-12-02+00-00",
        "tests/testcruise_evt/2014_185/2014-07-04T00-15-02+00-00.gz",
        "tests/testcruise_evt/2014_185/2014-07-04T00-17-02+00-00.gz",
        "tests/testcruise_evt/2014_185/2014-07-04T00-21-02+00-00",
        "tests/testcruise_evt/2014_185/2014-07-04T00-27-02+00-00"
    ]
    sfl_df = pd.read_parquet("tests/sfl.parquet")
    files_df = sfp.seaflowfile.date_evt_files(evt_files, sfl_df)
    answer_df = pd.DataFrame({
        'date': {
            0: pd.Timestamp('2014-07-04 00:00:02+0000', tz='UTC'),
            1: pd.Timestamp('2014-07-04 00:03:02+0000', tz='UTC'),
            2: pd.Timestamp('2014-07-04 00:06:02+0000', tz='UTC'),
            3: pd.Timestamp('2014-07-04 00:09:02+0000', tz='UTC'),
            4: pd.Timestamp('2014-07-04 00:12:02+0000', tz='UTC'),
            5: pd.Timestamp('2014-07-04 00:15:02+0000', tz='UTC'),
            6: pd.Timestamp('2014-07-04 00:17:02+0000', tz='UTC')
        },
        'file_id': {
            0: '2014_185/2014-07-04T00-00-02+00-00',
            1: '2014_185/2014-07-04T00-03-02+00-00',
            2: '2014_185/2014-07-04T00-06-02+00-00',
            3: '2014_185/2014-07-04T00-09-02+00-00',
            4: '2014_185/2014-07-04T00-12-02+00-00',
            5: '2014_185/2014-07-04T00-15-02+00-00',
            6: '2014_185/2014-07-04T00-17-02+00-00'
        },
        'path': {
            0: 'tests/testcruise_evt/2014_185/2014-07-04T00-00-02+00-00',
            1: 'tests/testcruise_evt/2014_185/2014-07-04T00-03-02+00-00.gz',
            2: 'tests/testcruise_evt/2014_185/2014-07-04T00-06-02+00-00',
            3: 'tests/testcruise_evt/2014_185/2014-07-04T00-09-02+00-00',
            4: 'tests/testcruise_evt/2014_185/2014-07-04T00-12-02+00-00',
            5: 'tests/testcruise_evt/2014_185/2014-07-04T00-15-02+00-00.gz',
            6: 'tests/testcruise_evt/2014_185/2014-07-04T00-17-02+00-00.gz'}
        }
    )
    hash_result = pd.util.hash_pandas_object(files_df, index=False).sum()
    hash_answer = pd.util.hash_pandas_object(answer_df, index=False).sum()
    assert hash_result == hash_answer
